from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import List, Optional


@dataclass
class Contributor:
    name: str
    email: str
    count: int = -1
    image: Optional[str] = None
    key: Optional[str] = None


class ContributionsReader(ABC):
    @abstractmethod
    def get_contributors(self, file_path: Path) -> List[Contributor]:
        """Obtains the list of contributors for a file with the given path."""

    @abstractmethod
    def get_last_commit_date(self, file_path: Path) -> datetime:
        """Reads the last commit date of a file."""



import re
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Iterable, List, Tuple

from dateutil.parser import parse as parse_date

from neoteroi.contribs.domain import ContributionsReader, Contributor


class GitContributionsReader(ContributionsReader):

    _name_email_rx = re.compile(r"(?P<name>[\w\s]+)\s<(?P<email>[^\>]+)>")

    def _decode(self, value: bytes) -> str:
        try:
            return value.decode("utf8")
        except UnicodeDecodeError:
            return value.decode("ISO-8859-1")

    def _parse_name_and_email(self, name_and_email) -> Tuple[str, str]:
        match = self._name_email_rx.search(name_and_email)
        if match:
            name = match.groupdict()["name"]
            email = match.groupdict()["email"]
        else:
            name, email = ""
        return name, email

    def parse_committers(self, output: str) -> Iterable[Contributor]:
        for line in output.splitlines():
            count, name_and_email = line.split("\t")
            name, email = self._parse_name_and_email(name_and_email)
            yield Contributor(name, email, int(count))

    def get_contributors(self, file_path: Path) -> List[Contributor]:
        """
        Obtains the list of contributors for a file with the given path,
        using the Git CLI.
        """
        in_process = subprocess.Popen(
            ["git", "log", "--pretty=short", str(file_path)], stdout=subprocess.PIPE
        )
        result = self._decode(
            subprocess.check_output(
                [
                    "git",
                    "shortlog",
                    "--summary",
                    "--numbered",
                    "--email",
                ],
                stdin=in_process.stdout,
            )
        )

        return list(self.parse_committers(result))

    def get_last_commit_date(self, file_path: Path) -> datetime:
        """Reads the last commit on a file."""
        result = self._decode(
            subprocess.check_output(
                ["git", "log", "-1", "--pretty=format:%ci", str(file_path)]
            )
        )
        return parse_date(result)



"""
This module provide a plugin that can be used to generate last modified time and
contributors list for each page, assuming that:

- the MkDocs site is built in a Git repository.
- the Git CLI can be used during the build.

"""
import logging
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from mkdocs.config import config_options as c
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File
from mkdocs.structure.pages import Page

from neoteroi.contribs.domain import ContributionsReader, Contributor
from neoteroi.contribs.git import GitContributionsReader
from neoteroi.contribs.html import ContribsViewOptions, render_contribution_stats

logger = logging.getLogger("MARKDOWN")


class ContribsPlugin(BasePlugin):
    _contribs_reader: ContributionsReader
    config_scheme = (
        ("contributors_label", c.Type(str, default="Contributors")),
        ("last_modified_label", c.Type(str, default="Last modified on")),
        ("time_format", c.Type(str, default="%Y-%m-%d %H:%M:%S")),
        ("contributors", c.Type(list, default=[])),
        ("show_last_modified_time", c.Type(bool, default=True)),
        ("show_contributors_title", c.Type(bool, default=False)),
    )

    def __init__(self) -> None:
        super().__init__()
        self._contribs_reader = GitContributionsReader()

    def _read_contributor_merge_with(self, contributor_info) -> Optional[str]:
        return contributor_info.get("merge_with")

    def _handle_merge_contributor_info(
        self,
        contributors: List[Contributor],
        contributor: Contributor,
        contributor_info: dict,
    ) -> bool:
        """
        Handles an optional "merge_with" property in the contributor info object
        (from configuration), returning true if the given contributor (from repo
        information) should be discarded (its commits count was merged with another
        one).
        """
        assert contributor.email == contributor_info.get(
            "email"
        ), "The contributor info object must match the contributor object."
        merge_with = contributor_info.get("merge_with")

        if merge_with:
            parent = next(
                (item for item in contributors if item.email == merge_with),
                None,
            )
            if parent:
                parent.count += contributor.count
                return True
        return False

    def _get_contributors(self, page_file: File) -> List[Contributor]:
        results = []
        contributors = self._contribs_reader.get_contributors(
            Path("docs") / page_file.src_path
        )
        info = self.config.get("contributors")

        if not info:
            return contributors

        for contributor in contributors:
            contributor_info = next(
                (item for item in info if item.get("email") == contributor.email), None
            )
            if contributor_info:
                contributor.image = contributor_info.get("image")
                contributor.key = contributor_info.get("key")

                if contributor_info.get("ignore"):
                    # ignore the contributor's information (can be useful for bots)
                    continue

                # should contributor information be merged with another object?
                if self._handle_merge_contributor_info(
                    contributors, contributor, contributor_info
                ):
                    # skip this item as it was merged with another one
                    continue

            results.append(contributor)

        return results

    def _get_last_commit_date(self, page_file: File) -> datetime:
        return self._contribs_reader.get_last_commit_date(
            Path("docs") / page_file.src_path
        )

    def _set_contributors(self, markdown: str, page: Page) -> str:
        page_file = page.file
        last_commit_date = self._get_last_commit_date(page_file)
        contributors = self._get_contributors(page_file)
        return markdown + render_contribution_stats(
            contributors,
            last_commit_date,
            ContribsViewOptions(
                self.config["contributors_label"],
                self.config["last_modified_label"],
                self.config["show_last_modified_time"],
                self.config["show_contributors_title"],
                self.config["time_format"],
            ),
        )

    def on_page_markdown(self, markdown, *args, **kwargs):
        try:
            markdown = self._set_contributors(markdown, kwargs["page"])
        except ValueError:
            logger.error(
                "Failed to display contributors list for page: %s",
                kwargs["page"].title,
            )
            pass
        return markdown



"""
This module contains methods to render the contributions stats.
"""
import xml.etree.ElementTree as etree
from dataclasses import dataclass
from datetime import datetime
from typing import List
from xml.etree.ElementTree import tostring as xml_to_str

from neoteroi.contribs.domain import Contributor


def _get_initials(value: str) -> str:
    return "".join([x[0].upper() for x in value.split(" ")][:2])


@dataclass
class ContribsViewOptions:
    contributors_label: str
    last_modified_label: str
    show_last_modified_time: bool
    show_contributors_title: bool
    time_format: str


def contribution_stats_to_element(
    contributors: List[Contributor],
    last_commit_date: datetime,
    options: ContribsViewOptions,
) -> etree.Element:
    element = etree.Element("div", {"class": "nt-contribs"})

    if options.show_last_modified_time:
        last_modified_time = etree.SubElement(
            element,
            "p",
            {"class": "nt-mod-time"},
        )
        if options.last_modified_label:
            last_modified_time.text = (
                f"{options.last_modified_label}: "
                f"{last_commit_date.strftime(options.time_format)}"
            )
        else:
            last_modified_time.text = last_commit_date.strftime(options.time_format)

    if options.show_contributors_title:
        contributors_title = etree.SubElement(
            element, "p", {"class": "nt-contributors-title"}
        )
        contributors_title.text = f"{options.contributors_label} ({len(contributors)})"

    contributors_parent = etree.SubElement(element, "div", {"class": "nt-contributors"})

    for i, contributor in enumerate(
        sorted(contributors, key=lambda item: item.count, reverse=True)
    ):
        props = {
            "class": "nt-contributor"
            + (f" {contributor.key}" if contributor.key else ""),
            "title": (
                f"{contributor.name} <{contributor.email}> ({contributor.count})"
            ),
        }

        if contributor.image:
            props.update(
                {
                    "class": "nt-contributor image",
                    "style": ("background-image: " f"url('{contributor.image}')"),
                }
            )
        props["class"] += f" nt-group-{i}"
        contrib_el = etree.SubElement(contributors_parent, "div", props)

        if not contributor.image:
            # display initials
            initials_el = etree.SubElement(contrib_el, "span", {"class": "nt-initials"})
            initials_el.text = _get_initials(contributor.name)
        else:
            etree.SubElement(contrib_el, "span", {"class": "nt-initials"})

    return element


def render_contribution_stats(
    contributors: List[Contributor],
    last_commit_date: datetime,
    options: ContribsViewOptions,
) -> str:
    return xml_to_str(
        contribution_stats_to_element(contributors, last_commit_date, options)
    ).decode("utf8")
